/*
 * Firewall Control Protocol module
 *
 * $Id: fcp_mod.c,v 1.5 2003/ 12:23:53 jaime Rel $
 *
 * Copyright (C) 2001-2003 Fhg Fokus
 *
 * This file is part of ser, a free SIP server.
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include "../../sr_module.h"
#include "../../dprint.h"
#include "../../mem/mem.h"
#include "../tm/t_hooks.h"
#include "../tm/tm_load.h"
#include "../tm/h_table.h"
#include "../../locking.h"
#include "../../timer.h"

#include "../../parser/msg_parser.h"
#include "../../str.h"


#include "fcp_mod.h"
#include "fcp.h"
#include "fcp_pa.h"
#include "tcpClient.h"
#include "text_ops.h"
#include "parser.h"
#include "session.h"

struct tm_binds tmb;

static int mod_init( void );
static void destroy(void);
static int child_init(int rank);

static int replace_contact_f(struct sip_msg*, char*, char*);
static int replace_via_f(struct sip_msg*, char*, char*);
static int replace_sdp_f(struct sip_msg*, char*, char*);
static int fcp_enable_f(struct sip_msg*);

db_con_t* db_handle;   /* Database connection handle. For future use? */

/* ----- Parameter variables ----------- */

char *fcp_server="127.0.0.1";
int fcp_port=12345;
int fcp_timer=60;
int fcp_flag=0;
int replace_contact_flag=0;
int replace_via_flag=0;
int replace_sdp_flag=0;
int usesyslog=1;
/* Use nat functionality? 0 No, 1 Yes */
int usenat=0;
/* IP address of SIP proxy */
char *sip_proxy_ip="192.168.6.153";
/* External IP address of NAT/FW */
char *firewall_ip="192.168.16.80";
/* External port address of NAT/FW */
char *firewall_port="5060";

/* Flag to enable FCP. Allows to know if we create children or not */
int check_rules=0;
int check_interval=5;

/* flag which is needed for reporting missed calls */
int missed_flag=2;

/* report e2e ACKs too */
int report_ack = 1;

/* should early media replies (183) be logged ? default==no */
int early_media = 0;

/* should failed replies (>=3xx) be logged ? default==no */
int failed_transactions = 0;


pthread_t pchild;

/* ------------- Callback handlers --------------- */

static void fcp_onoutboundresp( struct cell* t,  struct sip_msg *msg,
	int code, void *param );
static void fcp_onack( struct cell* t,  struct sip_msg *msg,
	int code, void *param );
static void fcp_onoutbound( struct cell* t, struct sip_msg *msg,
	int code, void *param ) ;
static void fcp_onmissed(struct cell *t, struct sip_msg *reply,
	int code, void *param );
static void fcp_onend(struct cell *t, struct sip_msg *reply,
	int code, void *param );


/*
 * Exported functions
 */
static cmd_export_t cmds[] = {
	{"replace_contact", replace_contact_f, 0, 0, REQUEST_ROUTE},
	{"replace_sdp", replace_sdp_f, 0, 0, REQUEST_ROUTE},
	{"replace_via", replace_via_f, 0, 0, REQUEST_ROUTE},
	{"fcp_enable", (cmd_function)fcp_enable_f, 0, 0, REQUEST_ROUTE},
	{0, 0, 0, 0, 0}
};




static param_export_t params[]={
	{"fcp_server", STR_PARAM, &fcp_server},
	{"fcp_port", INT_PARAM, &fcp_port},
	{"fcp_timer", INT_PARAM, &fcp_timer},
	{"usesyslog", INT_PARAM, &usesyslog},
	{"usenat", INT_PARAM, &usenat},
	{"sip_proxy_ip", STR_PARAM, &sip_proxy_ip},
	{"firewall_ip", STR_PARAM, &firewall_ip},
	{"firewall_port", STR_PARAM, &firewall_port},
	{"check_interval", INT_PARAM, &check_interval},
	{0,0,0}
};


struct module_exports exports = {
	"fcp",
	cmds,
	params,

	mod_init, /* initialization module */
       	0, /* response function */
	destroy,   /* destroy function */
        0,  		/* oncancel function */
	child_init  /* per-child init function */
};

void check_rule_list( unsigned int ticks, void *param) {

  struct session *last, *tmp;
  int local_port, remote_port;
  int rc, i = 0;
  char buffer[512];


  tmp = NULL;

  printf("check_rule_list() ");  

  get_lock2(&sl->lock);

  last = sl->last_session;
  
  release_lock2(&sl->lock);

  printf("check_rule_list() Going through session list...\n");  


  for (tmp=last; tmp!=NULL && i<20; tmp=tmp->previous) {
    if(!last->callid.s) {
      printf("check_list: no call id present in session!\n");
    }
    else {
      tmp->fcp_timer = tmp->fcp_timer-check_interval;
      printf("check_list: callid length=%d//expiry=%d\n", tmp->callid.len, tmp->fcp_timer);
      if((tmp->fcp_timer < 2*check_interval) && (tmp->fcp_timer > -5)) {
	printf("Renewing rules in firewall for this session!\n");


	/* Check whether socket is open */


	if(sd<0) {
	  LOG(L_ERR, "rule_checker: Could not recover connection to fcp server. Trying again\n");
	  sd = openConnection(fcp_server, fcp_port);

	  if(sd<0) {
	    LOG(L_ERR, "rule_checker: Could not recover connection. Don't process message...\n");
	    goto error;
	  }
	}

	if ((local_port = string_to_int(tmp->local_sdp.port.s)) < 0) {
	  printf("rule_checker(): Problem converting local sdp port\n");
	  //	  continue;
	  goto error;
	}
	
	if ((remote_port = string_to_int(tmp->remote_sdp.port.s)) < 0) {
	  printf("rule_checker(): Problem converting remote sdp port");
	  //	  continue;
	  goto error;
	}
	
	printf("fcp_onoutboundresp() renewing rule for:\n \
		    src_ip:port = %s:%s\n \
		    dst_ip:port = %s:%s\n",\
	       tmp->local_sdp.ip.s,\
	       tmp->local_sdp.port.s,\
	       tmp->remote_sdp.ip.s,\
	       tmp->remote_sdp.port.s);
	
	rc = open_pinhole(sd, tmp->local_sdp.ip.s,
			   tmp->remote_sdp.ip.s,
			   local_port,
			   remote_port,
			   fcp_timer);

	if(rc<0) {
	  printf("Problems renewing rules!!\n");
	  goto error;
	}
	
	rc = read_reply(sd, buffer);
	
	if(rc<0) {
	  perror("error listening to connection ");
	  closeConnection(sd);
	  goto error;
	}

	/* 
	   If we reached here, rules were renewed successfully.
	   Refresh the timer from the session.

	*/

	tmp->fcp_timer = fcp_timer;
	printf("Setting fcp_timer to %d\n", fcp_timer);


      }


    }

    i++;

  } /* Fi for */

 error:
  release_lock2(&sl->lock);

  /* Do nothing. May be next time...*/
 
  return;

}


void *Child(void *arg) {

    /* Child process */

  struct session *s;

  printf("FCP checker up. Checking sd = %d\n", sd);

  //  for(;;) {

  //    sleep(5);
    pthread_mutex_lock(&mutx);
    s = sl->last_session;
    pthread_mutex_unlock(&mutx);

    if (s!=NULL) {
      printf("child. last_session call id: %s\n", s->callid.s);
    }
    else {
      printf("child. No last_session present\n");
    }
    //  }
    //  pthread_exit((void*) 0);
  pthread_exit(NULL);
}



static int mod_init( void )
{

	load_tm_f	load_tm;

	fprintf( stderr, "fcp - initializing\n");

	/* import the TM auto-loading function */
	if ( !(load_tm=(load_tm_f)find_export("load_tm", NO_SCRIPT, 0))) {
		LOG(L_ERR, "ERROR: fcp: mod_init: can't import load_tm\n");
		return -1;
	}
	/* let the auto-loading function load all TM stuff */
	if (load_tm( &tmb )==-1) return -1;

	/* register callbacks */
	if (tmb.register_tmcb( TMCB_REPLY, fcp_onoutboundresp, 0 /* empty param */ ) <= 0)
		return -1;
	if (tmb.register_tmcb( TMCB_E2EACK, fcp_onack, 0 /* empty param */ ) <=0 )
		return -1;
	if (tmb.register_tmcb( TMCB_REQUEST_OUT, fcp_onoutbound, 0 /* empty param */ ) <=0 )
		return -1;
	/*	if (tmb.register_tmcb( TMCB_ON_NEGATIVE, fcp_onmissed, 0 *//* empty param *//* ) <=0 )*/
	//		return -1;
	//       	if (tmb.register_tmcb( TMCB_END, fcp_onend, 0 /* empty param */ ) <=0 )
	//  return -1;


	sd = openConnection(fcp_server, fcp_port);

	if(sd<0) {
	  perror("cannot open connection");
	  closeConnection(sd);
	  goto error;
	}
	
	printf("Parent. sd = %d\n", sd);
	
	//	sl=pkg_malloc(sizeof(struct session_list)); 
	sl=shm_malloc(sizeof(struct session_list));
	
	if (sl == NULL){
	  printf("Error initialising session list. Error!\n");
	  goto error;
	}
	
	sl->num_sessions = 0;
	
	sl->last_session = NULL;
	
	
	/* Start rule checker timer. Will keep track
	   of firewall rules
	*/
	
	if(check_interval == 0) {
	  printf("check_rules not enabled. Skipping firewall rule check\n");
	  goto cont;
	}

	/**********TIMER EXPERIMENT***********/

	if(register_timer(check_rule_list, 0, check_interval)<0) {
	  printf("mod_init() Register timer failed!\n");
	  goto error;	  
	}
	
	printf("Registered rule checker\n");
	
 cont:
	return 0;
	
 error:
	return -1;
	
}

static int child_init(int rank)
{


  /************************************************
      FORKING EXPERIMENT

  //  int childpid;

  //  pipe(pdes);

  //  sl=shm_malloc(sizeof(struct session_list));
  sl=pkg_malloc(sizeof(struct session_list)); 

  if (lock_init(&sl->lock)==0){
    printf("Error initialising lock. Error!\n");
    goto error;
  }


  sl->num_sessions = 0;

  sl->last_session = NULL;
  

  if((childpid=fork())<0 )
    {
      DBG("FCP :child_init: error - cannot launch rule checker\n");
      exit(-1);
    }

  if(childpid == 0)
  {
    struct session *last;

    for (;;) {

      sleep(5);
      
      lock_get(&sl->lock);

      printf("Child. Got the lock!\n");

      last = sl->last_session;

      lock_release(&sl->lock);

      printf("Child. Released the lock!\n");

      if(last != NULL) {
      	printf("Last session Call Id: %s\n", sl->last_session->callid.s);
      } else {
      	printf("No session present. Wait 5 seconds...\n");
      }
      
    }
    exit(0);

  }

  //  close(pdes[0]);

  //  signal(SIGALRM, send_last_session);

  //  printf("Child process pid=%d\n", childpid);


  ****************************************************/

  
  /*****************************

  PTHREADS EXPERIMENT

  //  pthread_t pchild;
  pthread_attr_t attr;
  //  pthread_mutex_t mutx;


  pthread_mutex_init(&mutx, NULL);

  pthread_attr_init(&attr);
  //  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  

  if(pthread_create(&pchild, &attr, Child, NULL) != 0) {
    printf("Error creating pthread\n");
    goto error;
  }  

  printf("Parent process pid=%d\n", getpid());

  //  pthread_create(&pchild, &attr, Child, NULL);

  pthread_attr_destroy(&attr);

  ************************************/


  return 0;

}

static void destroy(void)
{

  struct session *tmp;

#ifdef SQL_FCP
    if (db_handle) db_close(db_handle);
#endif

    /* Close the connection to the fcp server */
    closeConnection(sd);

    /* Deallocate shared memory */

    //    lock_destroy(&sl->lock);
    //    lock_dealloc(&sl->lock);

    if(sl->last_session == NULL) {
      printf("destroy() No last session to deallocate\n");
      goto cont;
    }

    printf("Deallocating list of sessions...\n");

    for (tmp=sl->last_session; tmp!=NULL; tmp=tmp->previous) {
      if(delete_session(tmp) < 0) {
	printf("destroy() Error deleting session from session list!\n");
      }
    } /* Fi for */

 cont:
    if(sl) {
      lock_destroy(&sl);
      // lock_dealloc(&sl);
    }


    /****************************************

     For the Pthreads experiment

    if (sl)
      pkg_free(sl);
    pthread_mutex_destroy(&mutx);
    pthread_exit(NULL);

    *************************************/

}

static void fcp_onoutbound( struct cell* t, struct sip_msg *msg,
	int code, void *param )
{
  /* 
     An outbound request came. We do not know if the transaction 
     will complete.Therefore, in the case of only firewall, flag 
     the transaction for the time being.
     For NAT's, we must ask the NAT device what is the external IP 
     address and port that will be open. Then change the Via, Contact 
     fields in the SIP header, and the IP and media field in the SDP 
     header. Anything else? Then send the message to the destination.
  */

  int rc, len, local_port, remote_port;
  char buffer[512];
  char *buf, *ptr;

  struct session *tmp_session, *tmp_session2;

  /* Sanity check */

  if (!check_transaction_quadruple( msg )) {
    LOG(L_ERR,"ERROR:fcp_onoutbound bad message!\n");
    goto error;
  }


  DBG("\n%s\n", msg->buf);

  if (!fcp_flag) {
    DBG("FCP flag not set. Don't process...\n");
    return; 
  }

  /* Create a session struct to track the connection */

  tmp_session = create_session();
  
  if (tmp_session==NULL) {
    LOG(L_ERR, "fcp_onoutbound: Error creating session!!");
    goto error;
  }

  rc = parse_callid(msg, tmp_session);

  if (rc<0) {
    LOG(L_ERR, "Error while extracting media params: \n");
    goto error;
  }

  /*
    Keep info related to local contact
  */
  
  rc = parse_contact_info (msg, msg, tmp_session);
  
  if (rc<0) {
    LOG(L_ERR, "Error while parsing contact info \n");
    goto error;
  }

  /* Only append session if it is not in the list */
  
  if ((tmp_session2 = find_session(sl->last_session, &tmp_session->callid))!=NULL) {
    /* Session exists. Therefore process according to
       what we already know of it
    */
    goto fast_process;
    
  }

  //    pthread_mutex_lock(&mutx);
  
  printf("fcponoutbound() Session not found. Append to list...\n");
  
  sl->last_session = append_session(tmp_session, sl->last_session);

  //    pthread_mutex_unlock(&mutx);
  
  tmp_session2 = sl->last_session;


 fast_process:


  /***********************************
      Is it an outbound INVITE ? 
  **********************************/
  

  if (t->is_invite) {

    rc = parse_media_info(msg, msg, tmp_session2);
  
    if (rc<0) {
      LOG(L_ERR, "Error while extracting media params: \n");
      goto error;
    }


    if (usenat) {
      

      /* Check whether socket is open */

      if(sd<0) {
	LOG(L_ERR, "fcp_onoutbound: Could not recover connection to fcp server. Trying again\n");
	sd = openConnection(fcp_server, fcp_port);
	if(sd<0) {
	  LOG(L_ERR, "fcp_onoutbound: Could not recover connection. Don't process message...\n");
	  goto error;
	}
      }


      rc = get_nat(sd, tmp_session->local_sdp.ip.s,
		   string_to_int(tmp_session2->local_sdp.port.s),
		   string_to_int(tmp_session2->local_sdp.port.s)+1);
      
      if(rc<0) {
	perror("Error sending request to NAT device");
	closeConnection(sd);
	sd = openConnection(fcp_server, fcp_port);
	if(sd<0) {
	  LOG(L_ERR, "fcp_onoutbound: Could not recover connection. Don't process message...\n");
	  goto error;
	}	
      }
      
      rc = read_reply(sd, buffer);
      
      if(rc<0) {
	perror("Error listening to connection ");
	closeConnection(sd);
	goto error;
      }
      
      /* Keep info related to the FCP transaction */
    
      rc = parse_fcp_info(buffer, tmp_session2);
    
      if(rc<0) {
	perror("error listening to connection ");
	goto error;
      }


      /* 
	 Replace Contact private IP address and
	 port from the natted end point
      */


      if(replace_contact_flag>0) {

	rc = replace_contact(msg, tmp_session2);
      
	if(rc<0) {
	  perror("Error replacing contact ");
	  goto error;
	}
      }

      /* 
	 Replace SDP private IP addresses and
	 ports from natted end points
      */

      if(replace_sdp_flag>0) {

	rc = replace_sdp(msg, tmp_session2);
	
	if(rc<0) {
	  perror("Error replacing sdp ");
	  goto error;
	}
      }

      /*
	Adds "rport" to via so that next proxy
	fills in the port and IP address the
	request cam from
      */

      if(replace_via_flag) {

	rc = replace_via(msg, tmp_session2);
	
	if(rc<0) {
	  perror("Error replacing via ");
	  goto error;
	}
      }    

    } /* endif use_nat */    

  
  } /* Endif is_invite */

  else 
    { /* NOT an INVITE. Make changes in contact header */

      if (usenat) {

	/* Check if it is a CANCEL or BYE */
  
	if(is_method(msg->buf, "CANCEL", 1) || is_method(msg->buf, "BYE", 1)) {
	  printf("fcp_onoutbound: Message is BYE or CANCEL. Close Pinhole!\n");
	  
	  
	  /* Close the pinhole related to the session. Need 
	     to parse_parameters from BYE's and
	     CANCELs
	  */
	    
	  /**********************
             Close Pinhole here 
	  *********************/

	  /* Check whether socket is open */

	  if(sd<0) {
	    LOG(L_ERR, "fcp_onoutbound: Connection to fcp server down. Trying again\n");
	    sd = openConnection(fcp_server, fcp_port);
	    if(sd<0) {
	      LOG(L_ERR, "fcp_onoutbound: Could not recover connection. Don't process message...\n");
	      goto error;
	    }
	  }
	  
	  if ((local_port = string_to_int(tmp_session2->local_sdp.port.s)) < 0) {
	    printf("fcp_onoutbound(): Problem converting local sdp port");
	    goto error;
	  }

	  if ((remote_port = string_to_int(tmp_session2->remote_sdp.port.s)) < 0) {
	    printf("fcp_onoutbound(): Problem converting remote sdp port");
	    goto error;
	  }
	  
	  printf("fcp_onoutbound() closing pinhole for:\n \
		    src_ip:port = %s:%s\n \
		    dst_ip:port = %s:%s\n",\
		 tmp_session2->local_sdp.ip.s,\
		 tmp_session2->local_sdp.port.s,\
		 tmp_session2->remote_sdp.ip.s,\
		 tmp_session2->remote_sdp.port.s);
    

	  
	  rc = close_pinhole(sd, tmp_session2->local_sdp.ip.s,
			     tmp_session2->remote_sdp.ip.s,
			     local_port,
			     remote_port);

	  if(rc < 0) {
	    printf("fcp_onoutboundresp(): Could not close pinhole\n");
	    goto error;
	  }

	  rc = read_reply(sd, buffer);
	  
	  if(rc<0) {
	    perror("error listening to connection ");
	    closeConnection(sd);
	    goto error;
	  }

	  /************************************

	  Check for answer from FCP server here. If
	  positive, ports related to session are closed.
          If so, "destroy" the temporary session and "delete"
          the session that was allocated in the list.
	  If not, what to do?

	  ***************************************/

	  /*
	  if(destroy_session(tmp_session)<0) {
	    printf("Cannot destroy session\n");
	    goto error;
	  }
	  */

	  if (delete_session(tmp_session2) == NULL) {
	    printf("fcp_onoutbound: Error deleting session!!\n");
	    goto error;
	  }

	  goto error;
	}

	printf("fcp_onoutbound: Not a CANCEL or BYE\n");
      
	/************************************* 
	 Replace Contact to point to NAT device
	****************************************/

	/* Create fw_ip/fw_port info */
	
	len = 3 /* IP= */ + strlen(firewall_ip) + 1 /* " " */ + 5 /* "PORT=" */ + strlen(firewall_port) + 1 /* " " */ + 5 /* "CSEQ=" */ + 4 /* seq++ */;
	
	buf = pkg_malloc(len + 1);
	
	if(!buf) {
	  printf("fcp_onoutbound(): malloc error\n");
	  goto error;
	}
	
	ptr = buf;
	
	append_str(ptr, "IP=", 3);
	append_str(ptr, firewall_ip, strlen(firewall_ip));
	append_str(ptr, " ", 1);
	
	append_str(ptr, "PORT=", 5);
	append_str(ptr, firewall_port, strlen(firewall_port));
	append_str(ptr, " ", 1);
	
	append_str(ptr, "CSEQ=", 4);
	append_str(ptr, "1234", 4);
	
	ptr = '\0';
	
	if (parse_fcp_info(buf, tmp_session2)<0) {
	  printf("fcp_onoutbound() Error parsing fcp info\n");
	  goto error;
	}
	
	DBG("NOT an INVITE. Replacing contact...\n");
	if(replace_contact_flag>0) {
	  
	  rc = replace_contact(msg, tmp_session2);

	  DBG("fcponoutbound: replace_contact rc= %d\n", rc);
	  
	  if(rc<0) {
	    perror("Error replacing contact ");
	    goto error;
	  }
	}

      }

    }

    /* setflag to recognise a flagged INVITE in subsequent responses */

    DBG("DEBUG: Outgoing request. Setting fcp flag... \n");
    setflag(msg, fcp_flag);


error:
	//if (buf) pkg_free(buf);
  return;
}

static void fcp_onmissed(struct cell *t, struct sip_msg *reply,
	int code, void *param )
{
  //	struct sip_msg *rq;

  //	rq = t->uas.request;

  //	if (t->is_invite
  //			&& missed_flag
  //			&& isflagset( rq, missed_flag)==1
  //			&& code>=300 )
  //	{
  //		fcp_missed_report( t, reply, code);
		/* don't come here again on next failed branch */
  //		resetflag(rq, missed_flag );
  //	}

/* A negative response came, like a 400 (Not Found, Unauthorised...).
   We must close any pinhole associated with that transaction. If a 
   300 redirect is received, need to get the Contact field and initiate
   a new INVITE to the new address */
}


/**********************

 Process responses from remote en points 

*************************/

static void fcp_onoutboundresp( struct cell* t, struct sip_msg *reply,
	int code, void *param )
{

  struct sip_msg *rq;
  struct session *tmp_session2, *tmp_session;

  int res, local_port, remote_port;
  
  int rc = 0;
  char buffer[512];

  /* Get the associated request */

  rq = NULL;
  rq = t->uas.request;
  if(rq->buf==NULL) {
    DBG("fcponoutboundresp: No request associated to this message. Discard..\n");
    goto error;
  } 

  /* Print message for debugging purposes */
  
  if (reply==FAKED_REPLY) {
    LOG(L_ERR, "fcponoutboundresp: FAKED_REPLY!!\n");
    goto error;
  }
    
  DBG("\n%s\n", reply->buf);
  
  tmp_session = NULL;
  tmp_session2 = NULL;
  
  /* fcp_onreply is bound to TMCB_REPLY which may be called
     we should not miss this event for missed calls either
  */

//	fcp_onmissed(t, reply, code, param );

  /* if fcp enabled for flagged transaction, check if flag matches */

  if (fcp_flag && isflagset( rq, fcp_flag ))
    DBG("fcponoutboundresp: Transaction previously flagged!!!\n");

  /* early media is reported only if explicitely demanded,
     other provisional responses are always dropped  */
  if (code < 200 && ! (early_media && code==183)) {
    DBG("fcponoutboundresp: early media. Discard\n");
    goto error;
  }

  /* negative transactions reported only if explicitely demanded */

  if (!failed_transactions && code >=300) {
    DBG("fcponoutboundresp: failed transaction...\n");    
    goto error;
  }
  
  /* anything else, i.e., 2xx, always reported */
  //	fcp_reply_report(t, reply, code);
  
  /* 
     If 200 response came from the remote end with SDP, we must open 
     the media port contained in the SDP media header. Ignore the 
     following ACK from the local end point 
  */

  /* Sanity check */

  if (!check_transaction_quadruple( reply )) {
    DBG("fcponoutbound Bad Message!\n");
    LOG(L_ERR,"ERROR:fcp_onoutbound bad message!\n");
    goto error;
  }

  tmp_session = create_session();

  if (tmp_session==NULL) {
    LOG(L_ERR, "fcp_onoutboundresp: Error creating session!!\n");
    goto error;
  }

  res = parse_callid(reply, tmp_session);

  if (res<0) {
    LOG(L_ERR, "fcp_onoutboundresp. Error while extracting callid: \n");
    goto error;
  }

  /*
    Keep info related to local and remote contact
  */
  
  res = parse_contact_info (rq, reply, tmp_session);
  
  if (res<0) {
    /* In 200 OK with SDP, Contact not always is present. Try at least */

    LOG(L_ERR, "fcp_onoutboundresp(): Error while parsing contact info\n");
    //    goto error;
  }


  /* Find the session in progress from the list. find_session_from_local
     will return NULL if the local_contact.ip and .port are not present.
     In that case, the session does not exist.
  */


  tmp_session2 = find_session(sl->last_session, &tmp_session->callid);

  if(tmp_session2 == NULL) {
    /* Session does not exist. Create it and append it to the list */

    //    pthread_mutex_lock(&mutx);
    
    sl->last_session = append_session(tmp_session, sl->last_session);

    //    pthread_mutex_unlock(&mutx);

    tmp_session2 = sl->last_session;


  } else {
    /* Session was in progress. Replace the one that
       was created for the outbound message with
       the complete information from the remote
       peer
    */

    //    pthread_mutex_lock(&mutx);

    sl->last_session = replace_session(tmp_session2, tmp_session);

    //    pthread_mutex_unlock(&mutx);

    /* last_session points to the last created session */

    if(sl->last_session == NULL) {
      LOG(L_ERR, "fcp_onoutboundresp: Error replacing session!!\n");
      goto error;
    }

    tmp_session2 = sl->last_session;
  }


  /* Check if it is a 200 OK */
  
  if(!is_method(reply->buf, "200", 0)) {
    printf("Message is not 200 OK. See what it is...\n");
    goto cont;	  
  }

  printf("Message is 200 OK...\n");  

  res = parse_media_info(rq, reply, tmp_session2);
  if (res<0) {
    printf("fcp_onoutboundresp. 200 without media params info\n");
    /* For the time being, don't process with this module */
    goto error;
  }

  
  /*
    Open the firewall for the media to flow. If there is
    nat, ask first to reserve ports
  */
  
  if (usenat) {
      
    /* 
       With the info from the NAT, open the ports for
       the media
    */

    /* Check whether socket is open */

    if(sd<0) {
      LOG(L_ERR, "fcp_onoutboundresp: Could not recover connection to fcp server. Trying again\n");
      sd = openConnection(fcp_server, fcp_port);
      if(sd<0) {
	LOG(L_ERR, "fcp_onoutboundresp: Could not recover connection. Don't process message...\n");
	goto error;
      }
    }

    if ((local_port = string_to_int(tmp_session2->local_sdp.port.s)) < 0) {
      printf("fcp_onoutboundresp(): Problem converting local sdp port");
      goto error;
    }

    if ((remote_port = string_to_int(tmp_session2->remote_sdp.port.s)) < 0) {
      printf("fcp_onoutboundresp(): Problem converting remote sdp port");
      goto error;
    }


    //    lock_get(&sl->lock);
   
    printf("fcp_onoutboundresp() opening pinhole for:\n \
		    src_ip:port = %s:%d\n \
		    dst_ip:port = %s:%d\n",\
		    tmp_session2->local_sdp.ip.s,\
		    local_port,\
		    tmp_session2->remote_sdp.ip.s,\
		    remote_port);
    
    res = open_pinhole(sd, tmp_session2->local_sdp.ip.s,
		       tmp_session2->remote_sdp.ip.s,
		       local_port,
		       remote_port,
		       fcp_timer);
  
    //    lock_release(&sl->lock);

    rc = read_reply(sd, buffer);
    
    if(rc<0) {
      perror("error listening to connection ");
      closeConnection(sd);
      goto error;
    }

    
  } /* endif(use_nat)*/
  
    /* 
       No NAT. Only open ports in firewall
    */
  else {

    printf("fcp_onoutboundresp() opening pinhole for:\n \
		    src_ip:port = %s:%s\n \
		    dst_ip:port = %s:%s\n",\
		    tmp_session2->local_sdp.ip.s,\
		    tmp_session2->local_sdp.port.s,\
		    tmp_session2->remote_sdp.ip.s,\
		    tmp_session2->remote_sdp.port.s);
    
    

    res = open_pinhole(sd, tmp_session2->remote_sdp.ip.s,
		       tmp_session2->local_sdp.ip.s,
		       string_to_int(sl->last_session->remote_sdp.port.s),
		       string_to_int(sl->last_session->local_sdp.port.s), fcp_timer);
    
    rc = read_reply(sd, buffer);
    
    if(rc<0) {
      perror("fcp_onoutboundresp: error listening to connection ");
      closeConnection(sd);
      goto error;
    }
  }

  /* Set the timer for the given session */

  lock_get(&sl->lock);

  tmp_session2->fcp_timer = fcp_timer;
  printf("fcp_ontouboundresp() New media session with timer set to %d\n", fcp_timer);

  lock_release(&sl->lock);

  goto error;


 cont:

  /****************************************************
    Check if it is a BYE/CANCEL for an existing session. In that
    case, close the pinhole associated with the session.

  *****************************************************/

  /* Check if it is a CANCEL or BYE */
  
  if(is_method(reply->buf, "CANCEL", 0) || is_method(reply->buf, "BYE", 0)) {
    printf("fcp_onoutboundresp: Message is BYE or CANCEL. Close Pinhole!\n");
    
    
    /* Close the pinhole related to the session. Need 
       to parse_parameters from BYE's and
       CANCELs
    */
    
    /**********************
      Close Pinhole here 
    *********************/

    /* Check whether socket is open */

    if(sd<0) {
      LOG(L_ERR, "fcp_onoutboundrespresp: Could not recover connection to fcp server. Trying again\n");
      sd = openConnection(fcp_server, fcp_port);
      if(sd<0) {
	LOG(L_ERR, "fcp_onoutboundresp: Could not recover connection. Don't process message...\n");
	goto error;
      }
    }
    
    if ((local_port = string_to_int(tmp_session2->local_sdp.port.s)) < 0) {
      printf("fcp_onoutboundresp(): Problem converting local sdp port");
      goto error;
    }

    if ((remote_port = string_to_int(tmp_session2->remote_sdp.port.s)) < 0) {
      printf("fcp_onoutboundresp(): Problem converting remote sdp port");
      goto error;
    }
    
    
    rc = close_pinhole(sd, tmp_session2->local_sdp.ip.s,
		       tmp_session2->remote_sdp.ip.s,
		       local_port,
		       remote_port);
    
    if(rc < 0) {
      printf("Could not close pinhole\n");
      goto error;
    }
    
    rc = read_reply(sd, buffer);
    
    if(rc<0) {
      perror("error listening to connection ");
      closeConnection(sd);
      goto error;
    }

    /************************************

     Check for answer from FCP server here. If
     positive, ports related to session are closed.
     If so, "destroy" the temporary session and "delete"
     the session that was allocated in the list.
     If not, what to do?

    ***************************************/

    if(destroy_session(tmp_session)<0) {
      printf("Cannot destroy session\n");
      goto error;
    }
	  
    if (delete_session(tmp_session2) == NULL) {
      printf("fcp_onoutbound: Error deleting session!!\n");
      goto error;
    }
    
    printf("fcp_onoutbound: Number of sessions: %d\n", sl->num_sessions);
    goto error;
  }



error:
	//if (buf) pkg_free(buf);
  return;

}


static void fcp_onack( struct cell* t , struct sip_msg *ack,
	int code, void *param )
{
	struct sip_msg *rq;

	rq = t->uas.request;
	/* only for those guys who insist on seeing ACKs as well */
	//	if (!report_ack) return;
	/* if fcp enabled for flagged transaction, check if flag matches */
	//	if (fcp_flag && isflagset( rq, fcp_flag )==-1) return;

	//	fcp_ack_report(t, ack);
}

static void fcp_onend( struct cell* t, struct sip_msg *msg,
		       int code, void *param ) {
  DBG("fcp_onend:\n%s\n", msg->buf);

}


static int replace_contact_f(struct sip_msg *msg, char* foo1, char * foo2) {

  //  setflag(msg, replace_contact_flag);
  replace_contact_flag=1;
  return 1;
}

static int replace_via_f(struct sip_msg *msg, char* foo1, char * foo2) {

  //  setflag(msg, replace_via_flag);
  replace_via_flag=1;
  return 1;
}

static int replace_sdp_f(struct sip_msg *msg, char* foo1, char * foo2) {

  //  setflag(msg, replace_sdp_flag);  
  replace_sdp_flag=1;
  return 1;
}

static int fcp_enable_f(struct sip_msg *msg) {

  //  setflag(msg, fcp_flag);  
  fcp_flag=4;
  return 1;
}

/*
void* Child2(void *arg){
*/
  /* Child process */
/*
  int tmp;
  int len, BUF_LEN;
  char buf[sizeof(struct session *) + 1];
  struct session *s;
  
  printf("FCP checker up. Checking sd = %d\n", sd);
*/  
  /* Close the writing pipe */
/*  close(pdes[1]);
  
  len = 0;
  
  printf("init_child: reading from pipe %d...\n", pdes[0]);
  
  for(;;) {
    
    sleep(5);
    
    kill(getppid(), SIGALRM);
    
    BUF_LEN = sizeof(struct session *);
    
    //      while(can_read(pdes[0], 30)) {
    
    //      while(len < sizeof(struct session)) {
    
    //	tmp = read(pdes[0], buf+len, sizeof(struct session));

    tmp = read(pdes[0], buf, BUF_LEN);
    
    switch (tmp) {
      
    case -1:
      printf("init_child(): read returned -1\n");
      close (pdes[0]);
      break;
    case 0:
      printf("init_child(): read returned 0\n");
      close (pdes[0]);
      break;
    case 4:
      printf("init_child(): read successful\n");
      break;
    default:
      break;
    }
    len+=tmp;
    //      }
    
    buf[len] = '\0';
    
    printf("child. last_session =\n%s\n", buf);
    
    s = (struct session *)buf;
    
    //      printf("child. last_session call id: %s\n", s->callid.s);
    */
    /*
      for (tmp=sl->last_session; tmp!=NULL; tmp=tmp->previous) {
      if(gettimeofday(&t, NULL) < 0) {
      printf("fcp_checker: Error getting time of day\n");
      return;
      }
      printf("fcp_checker(): timeofday %ld\n", t.tv_sec);
      
      if(tmp->fcp_timer<5) {
      printf("fcp_checker: Refreshing rules...\n");
    */
    /* Here the code to refresh rules... */
    
    
    //    } 
    
    
/*  } *//* Fi for */
/*  
  exit(0);
} 
*/

