/*
 * $Id: session.c,v 1.5 2003/ 12:23:53 jaime Rel $
 *
 * Copyright (C) 2001-2003 Fhg Fokus
 *
 * This file is part of ser, a free SIP server.
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <string.h> /* Memset */
#include <unistd.h> /* For write */
#include <sys/time.h>
#include <sys/types.h> /* For kill() */
#include <signal.h> /* For kill() */
#include <errno.h>
#include "session.h"
#include "../../str.h"
#include "../../locking.h"



/* Create a new Session struct initialised to 0 */

struct session* create_session() {

  struct session *s;

  //  s = (struct session*) pkg_malloc(sizeof(struct session));
  s = (struct session *) shm_malloc(sizeof(struct session));

  if(s <0) {
    LOG(L_ERR, "add_session: malloc error!!");
    goto error;
  }
  memset(s, 0, sizeof(struct session));

  return s;

 error:
  return NULL;
}


/* Destroy session. Deallocate memory from non listed session */

int destroy_session(struct session *deleted) {
  if(deleted == NULL) {
    printf("destroy_session:  parse error\n");
    return -1;
  }

  //  pkg_free(deleted);
  shm_free(deleted);
  return 1;
}
 
/* 
   Find session out of the list of
   sessions by call id
*/

struct session* find_session(struct session *last, str *callid) {

  struct session *tmp;
  int cmp;
  int i = 0;
  
   if (callid->s == NULL) {
    printf("find_session: no parsed callid!\n");
    return NULL;
  }

  printf("find_session: Searching for callid=%s\n", callid->s);

  get_lock2(&sl->lock);

  for (tmp=last; tmp!=NULL && i<20; tmp=tmp->previous) {
    if(!tmp->callid.s) {
      printf("find_session: no call id present in session!\n");
    }
    else {
      printf("find_session: comparing with callid=%s\n", tmp->callid.s);
      if((cmp=strncmp(tmp->callid.s, callid->s, callid->len))==0) {

	  /* This is the session we want */
	  printf("find_session: found session!\n");
       	  release_lock2(&sl->lock);
       	  return tmp;
      } /* Fi if */
    }
    i++;
  } /* Fi for */

  release_lock2(&sl->lock);

  /* Not found any */
  printf("find_session: No session found with call id %s\n", callid->s);
  return NULL;

}

/* Find a session out of the list of 
   local ip addresses and ports with active
   connections
*/

struct session* find_session_from_local (struct session *last, str *ip, str *port) {

  struct session *tmp;

   if (ip->s == NULL || port->s == NULL) {
    printf("find_session_from_local: no parsed port or ip!\n");
    return NULL;
  }

  printf("find_session_from_local: Searching for ip=%s, port=%s\n", ip->s, port->s);
  for (tmp=last; tmp!=NULL; tmp=tmp->previous) {
    if(!tmp->local_sdp.ip.s || !tmp->local_sdp.port.s) {
      printf("find_session_from_local: no local sdp present in session!\n");
    }
    else {
      printf("find_session_from_local: comparing with ip=%s, port=%s\n", tmp->local_sdp.ip.s, tmp->local_sdp.port.s);
      if(strncmp(tmp->local_sdp.ip.s, ip->s, ip->len)==0) {

	/* IP address is the same. Compare the port */

	if(strncmp(tmp->local_sdp.port.s, port->s, port->len)==0) {
	  /* This is the session we want */
	  printf("find_session_from_local: Found session: ip: %s port: %s\n", ip->s, port->s);
	  return tmp;
	} 
      } /* Fi if */
    }
  } /* Fi for */

  /* Not found any */
  printf("find_session_from_local: No session found with ip %s and port %s\n", ip->s, port->s);
  return NULL;
}

/* Find a session out of the list of active ports in
   firewall
*/

struct session* find_session_from_fw (struct session *last, str *ip, str *port) {

  struct session *tmp;

  if(last == NULL || ip == NULL || port == NULL) {
    printf("find_session_from_fw: parse error!");
    return NULL;
  }

  for (tmp=last; tmp!=NULL; tmp=tmp->previous) {
    if(ip->len!=tmp->fw_ip1.len) {
      /* IP address is not the one we're after. Jump to next! */
    } else {
      if(strncmp(tmp->fw_ip1.s, ip->s, ip->len)==0) {

      /* IP address is the same. Compare the port */

	if(port->len!=tmp->fw_port1.len) {
	  /* Port is not the one we're after */
	} else {
	  if(strncmp(tmp->fw_port1.s, port->s, port->len)==0) {
	    /* This is the session we want */
	    printf("find_session_from_fw: Found session: ip: %s port: %s\n", ip->s, port->s);
	    return tmp;
	  } 
	} /* Fi else */

      } /* Fi if */

    } /* Fi else */

  } /* Fi for */

  /* Not found any */
  printf("find_session_from_fw: No session found with ip %s and port %s\n", ip->s, port->s);
  return NULL;
}


/* Find a session out of the list of 
   remote ip addresses and ports with active
   connections
*/

struct session* find_session_from_remote (struct session *last, str *ip, str *port) {

  struct session *tmp;

  if(last == NULL || ip == NULL || port == NULL) {
    printf("find_session_from_remote: parse error!");
    return NULL;
  }

  for (tmp=last; tmp!=NULL; tmp=tmp->previous) {
    if(ip->len!=tmp->remote_sdp.ip.len) {
      /* IP address is not the one we're after. Jump to next! */
    } else {
      if(strncmp(tmp->remote_sdp.ip.s, ip->s, ip->len)==0) {

      /* IP address is the same. Compare the port */

	if(port->len!=tmp->remote_sdp.port.len) {
	  /* Port is not the one we're after */
	} else {
	  if(strncmp(tmp->remote_sdp.port.s, port->s, port->len)==0) {
	    /* This is the session we want */
	    printf("find_Session_from_remote: Found session: ip: %s port: %s\n", ip->s, port->s);
	    return tmp;
	  } 
	} /* Fi else */

      } /* Fi if */

    } /* Fi else */

  } /* Fi for */

  /* Not found any */
  printf("find_session_from_remote: No session found with ip %s and port %s\n", ip->s, port->s);
  return NULL;
}



/* Append a session to the list. Returns the pointer to
   the last session (the previous added) or NULL if an 
   error ocurred.
 */

struct session* append_session (struct session *s, struct session *last_s) {

  printf("append_session() ");

  get_lock2(&sl->lock);

  if(s == NULL) {
    printf("append_session: No session parsed!. Num sessions: %d", sl->num_sessions);
    goto error;
  }


  if(last_s == NULL) {
    /* First in the list */
    printf("append_session: First session in the list!!\n");
    s->next = NULL;
    s->previous = NULL;
    sl->num_sessions++;
    goto success;
  }

  printf("append_session: Not the first session. Adding...\n");
  last_s->next=s;
  s->previous = last_s;
  s->next = NULL;
  sl->num_sessions++;
  //  return s->previous;
  goto success;

 error:
  release_lock2(&sl->lock);
  return NULL;

 success: 
  printf("append_session: Num sessions: %d\n", sl->num_sessions);
  release_lock2(&sl->lock);
  return s;

  
}

/* Delete a session from the list. Returns the pointer
   to the last session in the list. NULL if error.
*/

struct session* delete_session(struct session *s) {

  struct session *s2, *last, *tmp;

  lock_get(&sl->lock);

  if(s==NULL) {
    printf("delete_session: No session parsed!. Num sessions: %d\n", sl->num_sessions);
    lock_release(&sl->lock);
    return NULL;
  }

  if(s->previous==NULL) {
    /* First in the list: just update the next's previous 
     and free memory */
    printf("delete_session: First session in the list!!\n");
    s2 = s->next;
    if(s2 == NULL) {
      printf("delete_session: First and last in the list!\n");
      sl->num_sessions--;
      printf("delete_session: Num sessions: %d!\n", sl->num_sessions);
      release_lock2(&sl->lock);
      //      return s;
      return NULL;
    }
    if (s2->previous != NULL) s2->previous = NULL;
    //    pkg_free(s);
    shm_free(s);
    sl->num_sessions--;
    goto success;
  }

  else if(s->next==NULL) {
    /* Last in the list; update the previous' next and
     free memory */
    printf("delete_session: Last session in the list!!\n");
    s2 = s->previous;
    s->previous->next = NULL;
    //    pkg_free(s);
    shm_free(s);
    sl->num_sessions--;
    goto success;
  }

  else {
    /* Must be in the middle of two sessions in the list */
    printf("delete_session: not first nor last session in the list!!\n");
    s->previous->next=s->next->previous;
    s2 = s->previous;
    //    pkg_free(s);
    shm_free(s);
    sl->num_sessions--;
    goto success;
  }

 success:

  for (tmp=s2; tmp!=NULL; tmp=s2->next)
    last = tmp;

  printf("delete_session: Num sessions: %d!\n", sl->num_sessions);
  release_lock2(&sl->lock);
  return s2;

}

/* Deletes old and appends new to the list 
   Returns the pointer to the new added session.Or NULL if error.
 */

struct session* replace_session (struct session *old, struct session *new) {

  struct session *tmp, *last;

  if(old == NULL) {
    printf("replace_session: Parse Error\n");
    lock_release(&sl->lock);
    return NULL;
  }

  /* Get the last session in the list */
  /*
  for (tmp=old; tmp!=NULL; tmp=old->next)
    last = tmp;
  */

  /* Delete the old session */

  if((last = delete_session(old)) == NULL) {
    printf("replace_session: Error deleting session\n");
    //    return NULL;
  }

  /* Append the new session */

  if((last = append_session(new, last)) == NULL) {
    printf("replace_session: could not append session\n");
    return NULL;
  }
  return last;
}

void send_last_session (struct session *last) {

  /* Send the last session */

  int len, tmp, w;

  printf("SIGALRM arrived. Processing...\n");

  if(!last) {
    printf("send_last_session(): No session active! Nothing to do...\n");
    return;
  }

  //  len = sizeof(struct session);
  len = sizeof(struct session);

  w = 0;
  
  /* Continue here. Send a known buffer!!!!! */

  //  while(w != len) {
    //    tmp = write(pdes[1], last + w, len - w);
    tmp = write(pdes[1], last, len);
    if (tmp == -1) {
      LOG(L_ERR, "send_last_session(): %s\n", strerror(errno));
      return;
    }
    
    w += tmp;

    //  }

    if (w != len) {
      printf("send_last_session(): %s\n", strerror(errno));
      return;
    }

  //  printf("send_last_session(): Sent:\n%s \n", last);
  printf("send_last_session(): %d bytes sent\n", w);

  
}
