/*
 * $Id: text_ops.c,v 1.5 2003/ 12:23:53 jaime Rel $
 *
 * Copyright (C) 2001-2003 Fhg Fokus
 *
 * This file is part of ser, a free SIP server.
 *
 * ser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version
 *
 * For a license to use the ser software under conditions
 * other than those described here, or to purchase support for this
 * software, please contact iptel.org by e-mail at the following addresses:
 *    info@iptel.org
 *
 * ser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <stdio.h>
#include <fcntl.h>

#include "text_ops.h"
#include "../../mem/mem.h"
#include "../../data_lump_rpl.h"
#include "../../locking.h"


/*********************** match ***************************
 
   Match a substring <pattern> within a string <string>.
   It will return the offset to the beginning of <pattern>
   in pm->rm_so, and the offset to the end of <pattern>
   in pm->rm_eo. string + pm->rm_eo is the address to
   the beginning of <pattern>

**********************************************************/

int match(const char *string, char *pattern, regmatch_t *pm)
{
    int    status;
    regex_t    re;

    if (regcomp(&re, pattern, REG_EXTENDED) != 0) {
        return(-1);      /* Report error. */
    }
    status = regexec(&re, string, (size_t) 1, pm, 0);
    regfree(&re);
    if (status != 0) {
        return(-1);      /* Report error. */
    }
    return(0); /* Successful */
}

/* 
   Convert string of chars (NULL terminated) to integer
*/

long int string_to_int(char *s) {

  if(s == NULL) {
    printf("string_to_int(): parse error!\n");
    return -1;
  }
  return strtol(s, (char **)NULL, 10);

}

/* 
   Convert int to char. Return number of
   characters printed. -1 if error ocurred.
   "dest" must be an already created buffer!
*/

int int_to_char(char *dest, int src, int size) {

  if(!dest) {
    printf("int_to_char(): parse error!\n");
    return -1;
  }

  return snprintf(dest, size, "%d", src);

}

/*
  Returns the offset to the end of the first appearance
  of <key> within <string>
*/

int get_offset_end(char *string, char *key) {

  regmatch_t pmatch;
  int status;

  status = match(string, key, &pmatch);
  if (status != 0) {
    return(-1);      /* Report error. */
  } else return pmatch.rm_eo;
}

/*
  Returns the offset to the beginning of the first appearance
  of <key> within <string>
*/

int get_offset_start(char *string, char *key) {

  regmatch_t pmatch;
  int status;

  status = match(string, key, &pmatch);
  if (status != 0) {
    return(-1);      /* Report error. */
  } else return pmatch.rm_so;
}

/* 
   Jumps characters till next delim
*/

char* skip_chars(char* _s)
{
  while ((*_s != ' ') && (*_s != '\r') && \
	 (*_s != '\n') && (*_s != '\t') && \
	 (*_s != ';') && (*_s != '\0') && \
	 (*_s !='-') && (*_s !='=') && \
	 (*_s !='<') && (*_s !='>') && \
	 (*_s !=':')) _s++;
  return _s;
}

/*
  Jumps characters till next word
*/

char* skip_delims(char* _s) {
  while (((*_s == ' ') || (*_s == '\r') || \
	 (*_s == '\n') || (*_s == '\t') || \
	 (*_s == ';') || (*_s == '-') || \
	 (*_s =='<') || (*_s =='>') || \
	 (*_s ==':') ||	(*_s == '=')) && \
	 (*_s != '\0')) _s++;
  return _s;
}


/* 
   Gets the value pointed by <key> from a string <s> and
   parses it as char* to <dest>. Returns the value of <key>.
   It expects s to contain
   key<delim1>...<delimn>value, where <delim> is {":", ";", 
   " ", ">", "<", "-", "\n", "\t", "\r" }
*/
 
int get_int_value(char *s, char *key) {


  char *buf, *ptr;
  int len;
  int end;

  char *string_end;
  char *string_start;
  

  end = get_offset_end(s, key);
  if(end < 0) {
    LOG(L_ERR, "SRC Port not found\n");
    return -1;
  }
  string_start = skip_delims(s + end);
  string_end = skip_chars(string_start);
	
  len = string_end - string_start;
  
  buf = shm_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:extcmd_add_contact: out of memory! \n");
    return -1;
  }
  
  ptr = buf;
  memcpy( ptr, string_start , len);
  *(ptr + len) = '\0';
  return string_to_int(buf);

}

/* 
   Gets the value pointed by <key> from a string <s> and
   parses it as char* to <dest>. Returns the
   length of the value of <key>. It expects s to contain
   key<delim>...<delimn>value, where <delim> is {":", ";",
   " ", ">", "<", "-", "\n", "\t", "\r" }

*/
 
char* get_char_value(char *s, char *key) {

  char *buf;
  int len;
  int end;

  char *string_end;
  char *string_start;
  
  end = get_offset_end(s, key);
  if(end < 0) {
    LOG(L_ERR, "SRC IP not found\n");
    return NULL;
  }
  string_start = skip_delims(s + end);
  string_end = skip_chars(string_start);
	
  len = string_end - string_start;
  
  buf = shm_malloc( len+1 ); /* extra character for \0? */
  if(!buf) {
    LOG(L_ERR,"ERROR:extcmd_add_contact: out of memory! \n");
    return NULL;
  }
  

  DBG("get_char_value: len = %d\n", len);

  memcpy( buf, string_start , len);
  *(buf + len) = '\0';
  return buf;

}

